/*
 * (C) Copyright Ian D. Romanick 2007
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * on the rights to use, copy, modify, merge, publish, distribute, sub
 * license, and/or sell copies of the Software, and to permit persons to whom
 * the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
 * AUTHORS, COPYRIGHT HOLDERS, AND/OR THEIR SUPPLIERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifdef _WIN32
# define WIN32_LEAN_AND_MEAN
# define _USE_MATH_DEFINES
# include <windows.h>

# pragma comment(lib, "SDL.lib")
# pragma comment(lib, "SDLmain.lib")
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glext.h>

#include <SDL/SDL.h>

#include "glsl_glue.h"
#include "vector.h"

extern bool extension_supported(const char *ext_name);
extern void generate_tangent_vectors(GLenum mode, const GLfloat *positions,
    const GLfloat *normals, GLfloat *tangents, unsigned count);


static bool done = false;
static bool anim = true;
static bool light_anim = true;
static SDL_Surface *my_surf = NULL;

static GLint t0 = 0;
static GLint frames = 0;
static GLfloat x_angle = 0.0f, y_angle = 0.0f;
static GLfloat light_angle = 0.0f;

idr_vec4 light_pos(1.0f, 1.0f, 5.0f, 1.0f);
GLhandleARB prog;

static GLint diffuse_color_uniform;
static GLint specular_color_uniform;
static GLint shininess;

static unsigned specular_idx = 1;
static unsigned diffuse_idx = 1;

static const GLfloat specular_colors[2][3] = {
    { 0.0, 0.0, 0.0 },
    { 0.9, 0.0, 0.0 }
};

static const GLfloat diffuse_colors[2][3] = {
    { 0.0, 0.0, 0.0 },
    { 0.7, 0.6, 0.18 }
};

static GLint bump_density;
static GLint bump_size;
static GLint light_pos_uniform;

static const GLfloat cube_verts[] = {
    1.0, 1.0, 1.0, 1.0,
    -1.0, 1.0, 1.0, 1.0,
    -1.0, -1.0, 1.0, 1.0,
    1.0, -1.0, 1.0, 1.0,

    1.0, 1.0, -1.0, 1.0,
    1.0, -1.0, -1.0, 1.0,
    -1.0, -1.0, -1.0, 1.0,
    -1.0, 1.0, -1.0, 1.0,

    1.0, 1.0, 1.0, 1.0,
    1.0, 1.0, -1.0, 1.0,
    -1.0, 1.0, -1.0, 1.0,
    -1.0, 1.0, 1.0, 1.0,

    1.0, -1.0, 1.0, 1.0,
    -1.0, -1.0, 1.0, 1.0,
    -1.0, -1.0, -1.0, 1.0,
    1.0, -1.0, -1.0, 1.0,

    1.0, 1.0, 1.0, 1.0,
    1.0, -1.0, 1.0, 1.0,
    1.0, -1.0, -1.0, 1.0,
    1.0, 1.0, -1.0, 1.0,

    -1.0, 1.0, 1.0, 1.0,
    -1.0, 1.0, -1.0, 1.0,
    -1.0, -1.0, -1.0, 1.0,
    -1.0, -1.0, 1.0, 1.0,
};


static const GLfloat cube_tex_coords[] = {
    0.0, 0.0,
    0.5, 0.0,
    0.5, 0.5,
    0.0, 0.5,

    0.0, 0.0,
    0.0, 1.0,
    1.0, 1.0,
    1.0, 0.0,

    0.0, 0.0,
    2.0, 0.0,
    2.0, 2.0,
    0.0, 2.0,

    0.0, 0.0,
    0.0, 0.25,
    0.25, 0.25,
    0.25, 0.0,

    0.0, 0.0,
    1.0, 0.0,
    1.0, 1.0,
    0.0, 1.0,

    0.0, 0.0,
    0.0, 0.5,
    1.0, 0.5,
    1.0, 0.0,
};


static const GLfloat cube_normals[24 * 3] = {
    0.0, 0.0, 1.0,
    0.0, 0.0, 1.0,
    0.0, 0.0, 1.0,
    0.0, 0.0, 1.0,

    0.0, 0.0, -1.0,
    0.0, 0.0, -1.0,
    0.0, 0.0, -1.0,
    0.0, 0.0, -1.0,

    0.0, 1.0, 0.0,
    0.0, 1.0, 0.0,
    0.0, 1.0, 0.0,
    0.0, 1.0, 0.0,

    0.0, -1.0, 0.0,
    0.0, -1.0, 0.0,
    0.0, -1.0, 0.0,
    0.0, -1.0, 0.0,

    1.0, 0.0, 0.0,
    1.0, 0.0, 0.0,
    1.0, 0.0, 0.0,
    1.0, 0.0, 0.0,

    -1.0, 0.0, 0.0,
    -1.0, 0.0, 0.0,
    -1.0, 0.0, 0.0,
    -1.0, 0.0, 0.0,
};

static GLfloat cube_tangents[24 * 3];


#define GL_CHECK_ERROR() \
    do { \
	GLenum err = glGetError(); \
	if (err) { \
	    printf("%s,%d: Error: 0x%04x\n", __FILE__, __LINE__, err); \
	} \
    } while (0)

#define GL_CHECK_ERROR_AND_LOG(h) \
    do { \
	GLenum err = glGetError(); \
	if (err) { \
	    GLsizei info_size; \
	    GLcharARB info_log[256]; \
	    printf("%s,%d: Error: 0x%04x\n", __FILE__, __LINE__, err); \
	    glGetInfoLogARB(h, sizeof(info_log), & info_size, info_log); \
	    if (info_size) { \
		printf("Info log:\n%s\n", info_log); \
	    } \
	} \
    } while (0)


static void Redisplay(void)
{
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


    glUseProgramObjectARB(0);

    glPointSize(3.0);
    glBegin(GL_POINTS);
    glColor3ub(0xff, 0xff, 0x00);
    glVertex3f(light_pos[0], light_pos[1], light_pos[2]);
    glEnd();


    glUseProgramObjectARB(prog);
    GL_CHECK_ERROR_AND_LOG(prog);

    glUniform3fARB(light_pos_uniform,
		   light_pos[0], light_pos[1], light_pos[2]);
    glUniform3fvARB(specular_color_uniform, 1, specular_colors[specular_idx]);
    glUniform3fvARB(diffuse_color_uniform, 1, diffuse_colors[diffuse_idx]);

    glPushMatrix();
    glRotatef(x_angle, 1.0f, 0.0f, 0.0f);
    glRotatef(y_angle, 0.0f, 1.0f, 0.0f);
    glDrawArrays(GL_QUADS, 0, 24);
    glPopMatrix();

    SDL_GL_SwapBuffers();
    frames++;

    GLuint t = SDL_GetTicks();
    if (t - t0 >= 5000) {
	GLfloat seconds = (GLfloat) (t - t0) / 1000.0f;
	GLfloat fps = frames / seconds;
	printf ("%d frames in %6.3f seconds = %6.3f FPS\n", frames, seconds, fps);
	t0 = t;
	frames = 0;
    }
}


static void Idle(void)
{
    static Uint32 last_t = ~0;
    Uint32 t = SDL_GetTicks();


    if (last_t != (Uint32)~0) {
	float dt = (float) (t - last_t) / 1000.0f;

	if (anim) {
	    x_angle += 45.0 * dt;
	    y_angle += 60.0 * dt;
	}

	if (light_anim) {
	    light_angle += 45.0 * dt;
	    light_pos[0] = sin(light_angle * M_PI / 180.0);
	}
    }

    last_t = t;
}


static void Init(void)
{
    if (!extension_supported("GL_ARB_fragment_shader")) {
	printf ("Sorry, this demo requires GL_ARB_fragment_shader\n");
	exit(1);
    }
    if (!extension_supported("GL_ARB_shader_objects")) {
	printf ("Sorry, this demo requires GL_ARB_shader_objects\n");
	exit(1);
    }
    if (!extension_supported("GL_ARB_shading_language_100")) {
	printf ("Sorry, this demo requires GL_ARB_shading_language_100\n");
	exit(1);
    }
    if (!extension_supported("GL_ARB_vertex_shader")) {
	printf ("Sorry, this demo requires GL_ARB_vertex_shader\n");
	exit(1);
    }

    init_glsl_glue();

    generate_tangent_vectors(GL_TRIANGLES, cube_verts, cube_normals,
			     cube_tangents, 24);

    /* Load, compile, link, etc. the vertex and fragment programs.
     */
    const GLcharARB *vertex_code = glsl_load_shader_code("vertex_shader.glsl");
    const GLcharARB *fragment_code = glsl_load_shader_code("fragment_shader.glsl");


    GLhandleARB vert_shader = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
    glShaderSourceARB(vert_shader, 1, &vertex_code, NULL);
    glCompileShaderARB(vert_shader);
    GL_CHECK_ERROR();
    delete vertex_code;

    GLhandleARB frag_shader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
    glShaderSourceARB(frag_shader, 1, &fragment_code, NULL);
    glCompileShaderARB(frag_shader);
    GL_CHECK_ERROR();
    delete fragment_code;

    prog = glCreateProgramObjectARB();
    glAttachObjectARB(prog, vert_shader);
    glAttachObjectARB(prog, frag_shader);
    glBindAttribLocationARB(prog, 1, "tangent");
    glLinkProgramARB(prog);
    GL_CHECK_ERROR_AND_LOG(prog);

    diffuse_color_uniform = glGetUniformLocationARB(prog, "diffuse_color");
    specular_color_uniform = glGetUniformLocationARB(prog, "specular_color");
    shininess = glGetUniformLocationARB(prog, "shininess");
    bump_density = glGetUniformLocationARB(prog, "bump_density");
    bump_size = glGetUniformLocationARB(prog, "bump_size");
    light_pos_uniform = glGetUniformLocationARB(prog, "light_pos");
    
    glUseProgramObjectARB(prog);
    GL_CHECK_ERROR_AND_LOG(prog);

    glUniform1fARB(shininess, 4.0);

    glUniform1fARB(bump_density, 6.0);
    glUniform1fARB(bump_size, 0.15);
    GL_CHECK_ERROR();



    glVertexPointer(4, GL_FLOAT, 0, cube_verts);
    glNormalPointer(GL_FLOAT, 0, cube_normals);
    glTexCoordPointer(2, GL_FLOAT, 0, cube_tex_coords);
    glVertexAttribPointerARB(1, 3, GL_FLOAT, GL_FALSE, 0, cube_tangents);

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableVertexAttribArrayARB(1);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);

    printf("GL_RENDERER = %s\n", (const char *) glGetString(GL_RENDERER));
    printf("Keyboard input:\n"
	   "    f: Toggle fullscreen.\n"
	   "    s: Toggle specular reflection.\n"
	   "    d: Toggle diffuse reflection.\n"
	   "    a: Toggle animation of cube.\n"
	   "    l: Toggle animation of light.\n"
	   "  ESC: Exit program.\n");
	   
}


static void Reshape(int width, int height)
{
    glViewport (0, 0, width, height);
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
    glFrustum (-1.0, 1.0, -1.0, 1.0, 5.0, 25.0);
    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();
    glTranslatef (0.0f, 0.0f, -15.0f);
}


static void Key(SDLKey sym)
{
    switch(sym) {
    case SDLK_ESCAPE:
	done = true;
	break;
                    
    case 'f':
	SDL_WM_ToggleFullScreen(my_surf);
	break;

    case ' ':
    case 'a':
	anim = !anim;
	break;

    case 'l':
	light_anim = !light_anim;
	break;

    case 's':
	/* This '1 - n' logic causes the variable, which is used as an array
	 * index to toggle between 0 and 1.
	 */
	specular_idx = 1 - specular_idx;
	break;

    case 'd':
	diffuse_idx = 1 - diffuse_idx;
	break;

    default:
	break;
    }
}


/**
 * Push an event each time the timer callback fires.
 */
Uint32 timer_callback(Uint32 interval, void *not_used)
{
    SDL_Event e;

    (void) not_used;

    e.type = SDL_USEREVENT;
    e.user.code = 0;
    e.user.data1 = NULL;
    e.user.data2 = NULL;
    SDL_PushEvent(& e);

    return interval;
}


int main (int argc, char *argv[])
{
    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0) {
	exit(1);
    }
    
    my_surf = SDL_SetVideoMode(800, 600, 0, SDL_OPENGL);
    if (my_surf == NULL) {
	SDL_Quit();
	exit(1);
    }

    Reshape(800, 600);
    Init();

    SDL_TimerID timer_id = SDL_AddTimer(10, timer_callback, NULL);

    while (!done) {
        SDL_Event event;

        SDL_WaitEvent(&event);
	do {
            switch (event.type) {
            case SDL_VIDEORESIZE:
                Reshape(event.resize.w, event.resize.h);
                break;

            case SDL_QUIT:
                done = true;
                break;
                
            case SDL_KEYDOWN:
		Key(event.key.keysym.sym);
                break;

            default:
                break;
            }
        } while (SDL_PollEvent(&event));


	Idle();
	Redisplay();
    }

    SDL_RemoveTimer(timer_id);

    SDL_Quit();
    return 0;
}
